//Krister Swenson                                                  Spring 2008
//
//Creates random permutations and tries one of the O(n lg(n)) methods on them.
//

#include <iostream>
#include "Permutation.h"
#include "../seqtools/seqfuncs.h"


#define REPS 100

//::ALGORITHM RELATED DEFS:
//Find the full sorting path even when encountering unsafe inversions:
//  (otherwise just fail and do restart depending on the value of RESTARTS)
#define FIX_UNSAFE

//Should be one of RANDOM, USE_MAX, or USE_MIN:
#define METHOD USE_MIN
//If one side is unsafe, then do the other:
// #define TRY_OTHER_ON_FAILURE


//::TEST SCRIPT RELATED DEFS:
// #define ENSURE_NO_ADJACENCIES  //Check permutations with only breakpoints.
// #define PRINT_P_ON_FAILURE     //Print the permutation when we don't sort it.
#define CONFIRM_SEQUENCE       //Confirm the computed reversal sequence.

#define RESTARTS1 1       //Try the chosen method this number of times.
#define RESTARTS2 0       //Try the chosen method AGAIN this number of times.
#define SEQ_LEN 100       //The length of the random permutations generated.







inline bool sort(Permutation &q)
  {
  vector<int> orig,qv;
  orig = qv = q.getSeqVec();
  list<ipair> revs;

  //cerr << "q:" << orig << endl;

#ifdef FIX_UNSAFE
  revs = q.getSortSeq(METHOD);
#ifdef CONFIRM_SEQUENCE
  for(list<ipair>::iterator i=revs.begin(); i!=revs.end(); i++)
    applyRev(qv,*i);
  if(getLinearTimeDist(qv) != 0)
    {
    cerr << "Reversal sequence failed for" << endl;
    cerr << "q:" << orig << endl;
    exit(0);
    }
#endif //CONFIRM_SEQUENCE
#else
  ipair rev;
  while(q.nextRev(rev, METHOD));
#endif


  if(q.isSorted())
    return true;
  else
    {
#ifdef TRY_INVERSE
    invertSeq(orig);
    Permutation p(orig);
    while(p.nextRev(rev, METHOD));
    if(p.isSorted())
      return false;     //The inverse permutation didn't get stuck!
    else
      return true;      //The inverse permutation got stuck too.
#endif
    //cerr << "q:" << orig << endl;
    return false;
    }
  }




int main()
  {
  int counter = 0, failure1 = 0;
  int failure2 = 0;

  cerr << "Sequence Length:" << SEQ_LEN << endl;
  srandom(time(NULL));    
  for(int i=0; i<REPS; i++)
  {
    vector<int> seq = getRandSPerm(SEQ_LEN);
    Permutation p(seq);
    if(p.hasHurdle())       //ensure no hurdles
      continue;
#ifdef ENSURE_NO_ADJACENCIES
    if(hasAdjacency(seq))   //ensure no adjacencies
      continue;
#endif

    bool val = false;
    for(int i = 0; i < RESTARTS1; i++)
      {
      Permutation q(seq);
      val = sort(q);
      if (val)
        break;
      }

    if(!val && RESTARTS1)
      {
      failure1++;
      //cerr << "Not sorted in " << RESTARTS << " trials." << endl;
#ifdef PRINT_P_ON_FAILURE
      cerr << "p:" << seq << endl;
#endif
      }
    //else
      //cerr << "Sorted in " << i+1 << " trials." << endl;

    val = false;
    for(int i = 0; i < RESTARTS2; i++)
      {
      Permutation q(seq);
      val = sort(q);
      if (val)
        break;
      }

    if(!val && RESTARTS2)
      {
      failure2++;
      //cerr << "Not sorted in " << RESTARTS << " trials." << endl;
#ifdef PRINT_P_ON_FAILURE
      cerr << "p:" << seq << endl;
#endif
      }
    //else
      //cerr << "Sorted in " << i+1 << " trials." << endl;


 
    counter++;
    cerr << endl;
    cerr << "Total: " << counter << " Failures(" << RESTARTS1 << "): "
         << failure1 << endl;
    cerr << "Total: " << counter << " Failures(" << RESTARTS2 << "): "
         << failure2 << endl;
    }
  return 0;
  }

