/*  Krister Swenson                                        Fall 2002
  Create two random permutation of the identity string by doing insertions,
reversals, and deletions.  One permutation is taken on the path from the
identity to the other.
*/


#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "rngmit.h"

#define unifRandVars 500

int LENGTH = 200;
int NUMOPS = 90;

//PROBABILITIES:
                    //The probability of seeing an operation for a given event:
int PrREV = 80;
int PrDEL = 10;
int PrINS = 10;
                    //The probability associated with insertions:
int PrCOPY = 50;   //The probability that an insert is a copy.

int NumDataElms = 1; //the number of perms to generate
int GAUSSMEAN = 7; //gaussian mean
int GAUSSSD = 2; //gaussian sd

void resize(int* perm, int len, int clen);
int gaussian(float mean, float sd);

//generates the sequences with the parameters described by the defines in
//this file we write to stdout the "Copies:" numCopies "Seq:" length the 
//generated sequence 
int genSeqs(int* S, int& actlen, int& size, int& nextval);

//this reads in a sequence from stdin return an array of it's elems put the
// len of that array in len and the number of cover elems affected by copies
// in numCopies
int* readInSeq(int* len, int* numCopies);



int main(int argc, char** argv)
  {
  cin >> LENGTH;
  cin >> NUMOPS;
  cin >> PrREV;
  cin >> PrDEL;
  cin >> PrINS;
  cin >> PrCOPY;
  cin >> NumDataElms;
  cin >> GAUSSMEAN;
  cin >> GAUSSSD;

  int size = LENGTH+(9*NUMOPS);    //The size of array S.
  int* S = new int[size];          //The sequence.
  int actlen = LENGTH;             //The ACTual LENgth of the string in S.
  int nextval = LENGTH+1;          //The smallest value that hasn't been used.
      
  rngseed((unsigned)time(NULL));   //Seed the random number generator.
  for(int numSeq = 0; numSeq < NumDataElms; ++numSeq)
    {
    for(int i=0; i < LENGTH; i++)  //Create the identity permutation:
      S[i] = i+1;

    genSeqs(S, actlen, size, nextval);
    }
  }


int genSeqs(int* S, int& actlen, int& size, int& nextval)
  {
  int numCopies = 0; //the number of cover elements that cover elements with
                     // multiple appearances in s
  /*
    cerr << "GAUSS: ";
    for(int i=0; i<20; i++)
    {
    cerr << gaussian(5.0, 1.5) << ",";
    }
    cerr << endl;
    return 1;
  */
    
  for(int k=0; k < NUMOPS; k++)
    { 
    unsigned int rnum = rngmitint; //Copy this for use in -insert- section.
    int Pr = rnum%100;
    int hold = rnum;
    rnum /= 100;
    if(Pr < PrREV)                 //DO A REVERSAL:
      {
      unsigned int start = rngmitint;    //Find the start and end genes.
      start = start%actlen;
      unsigned int end = rngmitint;
      end = start + (end%(actlen-start));

      //cerr << "--- PR:" << Pr << " hold:" << hold << endl;
      cerr << "REV:" << start << "," << end << endl;
      int temp;
      while(end >= start)
        {
        temp = -1 * S[end]; 
        S[end] = -1 * S[start];
        S[start] = temp;
        end--;
        start++;
        }
      }
    else if(Pr < PrREV+PrDEL)      //DO A DELETION:
      {
      int spot = abs(rngmitint) % actlen;
      int delSize = (int) abs(gaussian(GAUSSMEAN, GAUSSSD));
      if(delSize == 0) delSize++;
      delSize = delSize % (actlen - spot);

      cerr << "DEL:" << spot << " len:" << delSize << endl;
      
      while(spot+delSize < actlen)
        {
        S[spot] = S[spot + delSize];
        ++spot;
        }

      actlen = actlen - delSize;
      }
    else                           //DO AN INSERT:
      {
      Pr = rnum%100;
                                            //(We can insert in n+1 spots)
      unsigned int spot = rngmituint%(actlen+1);
      cerr << "INS(after):" << spot << " actualLEN:" << actlen << endl;
                                            //Number of genes to be inserted.
      int inslen = abs(gaussian(GAUSSMEAN, GAUSSSD));
      if(spot == 0)
        spot = 1;
      if(inslen == 0)
        inslen = 1;

      int ins[inslen];                      //The string to insert.

      cerr << "insLen:" << inslen << endl;

      if(Pr < PrCOPY)                       //-Make a copy:
        {
        unsigned int copspot = rngmituint%(actlen-inslen+1);
        ++numCopies;
        cerr << "   copspot:" << copspot << endl; 
        for(int i=0; i < inslen; i++)
          {
          ins[i] = S[i+copspot];
          cerr << "   " << ins[i] << " ";
          }
        cerr << "." << endl;
        }
      else                                  //-Add a new one:
        {
        cerr << "   new\n";
        for(int i=0; i < inslen; i++)
          ins[i] = nextval++;
        }

      if(actlen+inslen > size)              //Enlarge S if need be:
        {
        cerr << "RESIZE: " << actlen+inslen << "," << size << endl;
        size = actlen+inslen+NUMOPS;
        resize(S, size, actlen);
        }

      for(int i=actlen-1; i >= spot; i--)   //Make space.
        S[i+inslen] = S[i];
      for(int i=0; i < inslen; i++)         //Copy the elements.
        S[i+spot] = ins[i];

      actlen += inslen;
      }
    }

  cout << LENGTH << " " << 2*numCopies << " " << actlen << " ";
  for(int j=0; j < actlen; j++)
    cout << S[j] << " ";

  cout << endl;
  }


/* Enlarge perm to -len- and keep the leftmost -clen- elements.
*/
void resize(int* perm, int len, int clen)
  {
cerr << "RESIZE" << endl;
  int* temp = new int[len];
  for(int i=0; i < clen; i++)
    {
    temp[i] = perm[i];
    }

  delete(perm);
  perm = temp;
  }

//rngmitint to get a random int

int gaussian(float mean, float sd)
{
  long i, kForSigmaI, t1, t2;
  float unifSum, meanAdj, kForSigmaF, f1;

  //calculate the desired value of k to get the correct sigma
  kForSigmaF = (sqrt(12) * sd) + 1;
  meanAdj = kForSigmaF / 2.0;

  kForSigmaI = ((int) kForSigmaF) + 5;
  unifSum = 0;

  for(i = 0; i < unifRandVars; ++i)
    {
      f1 = rngmit;
      //cerr << mean - meanAdj << " " << fabs(f1) << endl;
      unifSum = unifSum + (fabs(f1) * kForSigmaF) - meanAdj;
    }
 
  //unifSum = unifSum - unifRandVars * meanAdj;
  //cerr << unifSum - unifRandVars * meanAdj << endl;
  return (int) ((unifSum / sqrt(unifRandVars)) + mean);
}

int* readInSeq(int* len, int* numCopies)
{
  int* input;
  int i;

  cin >> *numCopies;
  cin >> *len;

  input = new int[*len];

  for(i = 0; i < *len; ++i)
    cin >> input[i];
  
  return input;
}

