//Mark Marron, Krister Swenson                               Fall 2003
//This file is the third attempt at the multi-set of genomes renaming code.
// Hopefully it contains less stupidity than the first and second versions.
//
//  Renames two sequences containing copies.  The first one given is the source
//and the second is the target.  The original copy is chosen by repeatedly
//finding the largest contiguous sequence of copied elements in the target
//that matches the largest possible contiguous sequence in the source.
//   -- In the case of overlap length ties, shorter sequences win.
//

#include "renamemulti.h"
#define VERBOSE


using namespace std;
using namespace __gnu_cxx;

int mlistsize;

/* Pick the original for each copy existing in both sequences.  Rename the
them accordingly.
*/
void renameMulti(int* s1, int ls1, int* s2, int ls2)
  {
  int i;
  int numMatches;
  match* matches;

  //collapseGaps(s1, ls1, s2, ls2);   //Skip these because we no longer are
                                      //looking at contiguous sequences.

  if(ls1 <= ls2)                      //Ensure that the shorter sequence is
    {                                 //given as the first.
    matches = matchingLoop(s1, ls1, s2, ls2, &numMatches);
    doCopyRenaming(s1, s2, ls1, ls2, matches, numMatches);
    }
  else
    {
    matches = matchingLoop(s2, ls2, s1, ls1, &numMatches);
    doCopyRenaming(s2, s1, ls2, ls1, matches, numMatches);
    }

  for(i = 0; i < numMatches; ++i)
    freeMatch(matches[i]);
    
  free(matches);

  renameSingleSideCopies(s1, ls1);    //Rename the copies that aren't in s2.
  renameSingleSideCopies(s2, ls2);    //Rename the copies that aren't in s1.

#ifdef VERBOSE
  cerr << "Done with rename." << endl;
#endif // VERBOSE

  return;
  }



/* Pair up the duplicates. Return an array -A- where A[i] is the assignment
such that s1[i] = A[i].  A[i] < 0 if there is no corresponding gene.
   ** user resposible for the array memory returned.
*/
int* getDupArray(int* s1, int ls1, int* s2, int ls2)
  {
  int numMatches;
  match* matches;
  int* mapping = (int*) calloc(ls1, sizeof(int));
  for(int i=0; i<ls1; i++)
    mapping[i] = -1;

  if(ls1 <= ls2)     //Ensure that the shorter sequence is given as the first.
    matches = matchingLoop(s1, ls1, s2, ls2, &numMatches);
  else
    {
    matches = matchingLoop(s2, ls2, s1, ls1, &numMatches);
                     //Switch the values in the match so the naming is
                     // consistent:
    for(int i=0; i < numMatches; i++)
      {
      int tmp = matches[i]->indexS2;
      matches[i]->indexS2 = matches[i]->indexS1;
      matches[i]->indexS1 = tmp;
      }
    }
                     //Build the map -A- to return:
  for(int i=0; i < numMatches; i++)    //For each match,
    {                                  //map the indices...
    if(s1[matches[i]->indexS1] == s2[matches[i]->indexS2])
      for(int j=0; j < matches[i]->length; j++)
        mapping[matches[i]->indexS1 + j] = matches[i]->indexS2 + j;
    else
      for(int j=0; j < matches[i]->length; j++)
        mapping[matches[i]->indexS1 + j] = matches[i]->indexS2 +
                                           matches[i]->length - j - 1;
    }
 
  for(int i = 0; i < numMatches; i++)   //Free the matches.
    freeMatch(matches[i]);
  free(matches);

  return mapping;
  }


/* Pair up the duplicates. Return a vector of structures relating the elements.
*/
vector<DupInfo> getDupInfo(int* s1, int ls1, int* s2, int ls2)
  {
  int numMatches;
  match* matches;
  vector<DupInfo> mapping;

  if(ls1 <= ls2)     //Ensure that the shorter sequence is given as the first.
    matches = matchingLoop(s1, ls1, s2, ls2, &numMatches);
  else
    {
    matches = matchingLoop(s2, ls2, s1, ls1, &numMatches);
                     //Switch the values in the match so the naming is
                     // consistent:
    for(int i=0; i < numMatches; i++)
      {
      int tmp = matches[i]->indexS2;
      matches[i]->indexS2 = matches[i]->indexS1;
      matches[i]->indexS1 = tmp;
      }
    }
                     //Build the mapping to return:
  for(int i=0; i < numMatches; i++)    //For each match,
    {                                  //fill in the mapping...
    if(s1[matches[i]->indexS1] == s2[matches[i]->indexS2])
      for(int j=0; j < matches[i]->length; j++)
        mapping.push_back(DupInfo(abs(s1[matches[i]->indexS1 + j]),
                                  matches[i]->indexS1 + j,
                                  matches[i]->indexS2 + j,
                                  matches[i]->length));
    else
      for(int j=0; j < matches[i]->length; j++)
        mapping.push_back(DupInfo(abs(s1[matches[i]->indexS1 + j]),
                                  matches[i]->indexS1 + j,
                                  matches[i]->indexS2 + matches[i]->length -j-1,
                                  matches[i]->length));
    }
 
  for(int i = 0; i < numMatches; i++)   //Free the matches.
    freeMatch(matches[i]);
  free(matches);

  return mapping;
  }







/* Remove all genes that have duplicates from both sequences.
*/
void removeDups(int* s1, int &ls1, int* s2, int &ls2)
  {
  hash_map< int, int, hash<int> > beenseen1;//True if the value has been seen.
  hash_map< int, int, hash<int> > beenseen2;//True if the value has been seen.
  hash_map< int, int, hash<int> > doremap;  //True if the value has been seen.

  int i;
  for(i=0; i < ls1; i++)            //Mark elements that exist in s1:
    {
    if(beenseen1[abs(s1[i])])               //If we already saw the element
      doremap[abs(s1[i])] = 1;              // then mark it to be removed.
    else
      beenseen1[abs(s1[i])] = 1;            // otherwise, mark it as seen.
    }
  for(i=0; i < ls2; i++)            //Mark elements that exist in s2:
    {
    if(beenseen2[abs(s2[i])])
      doremap[abs(s2[i])] = 1;
    else
      beenseen2[abs(s2[i])] = 1;
    }


  int curind = 0;
  int newl1 = ls1;
  int newl2 = ls2;
  for(i=0; i < ls1; i++)
    {
    if(!doremap[abs(s1[i])])        //If we're not removing this one,
      s1[curind++] = s1[i];         //then move it to the current location.
    else
      newl1--;                      //Otherwise, decrement the length of l1.
    }

  curind = 0;
  for(i=0; i < ls2; i++)
    {
    if(!doremap[abs(s2[i])])
      s2[curind++] = s2[i];
    else
      newl2--;
    }

  cerr << "Shortened By: " << ls1-newl1 << "/" << ls1 << " and "
       << ls2-newl2 << "/" << ls2 << ".\n";
  ls1 = newl1;
  ls2 = newl2;
  }


//////////////////////////////////////////////////////////////////////////////////////////////
//definitions for the matches

/*Fill a match with the provided values.  If -isf- is false we are looking at a
reversed string and need to correct the index using -S1len-.
*/
void fillMatch(theMatch* mlist, int ind, int i1, int i2, int len,
               bool isf, int S1len)
  {
  assert(ind < mlistsize);       //Make sure we're not going out of bounds.
  if(isf)
    mlist[ind].indexS1 = i1;
  else
    mlist[ind].indexS1 = S1len - i1 - len;
  mlist[ind].indexS2 = i2;
  mlist[ind].length = len;
  mlist[ind].isforward = isf;
  }

match copyMatch(match mat)
  {
  match m;

  m = (match) malloc(sizeof(theMatch));

  m->length = mat->length;
  m->indexS1 = mat->indexS1;
  m->indexS2 = mat->indexS2;
  m->isforward = mat->isforward;
  m->overlaps = mat->overlaps;

  return m;
  }

match makeMatch(int len, int iS1, int iS2)
  {
  match m;

  m = (match) malloc(sizeof(theMatch));

  m->length = len;
  m->indexS1 = iS1;
  m->indexS2 = iS2;

  return m;
  }

void freeMatch(match m)
  {
  //fprintf(stderr, "renameMulti free2 %x\n", m);
  free(m);
  }

/*  Each match has two substrings associated with it.  This returns the number
of substrings that overlap each other.  There could be a total of 2 if the
matches overlap each other in both sequences (must be from opposite sides in
each).
*/
int matchOverlapCount(theMatch m1, theMatch m2)
  {
  int count = 0;

  int endm1s1 = m1.indexS1 + m1.length;
  int endm1s2 = m1.indexS2 + m1.length;
  int endm2s1 = m2.indexS1 + m2.length;
  int endm2s2 = m2.indexS2 + m2.length;
  int begm1s1 = m1.indexS1;
  int begm2s1 = m2.indexS1;
  int begm1s2 = m1.indexS2;
  int begm2s2 = m2.indexS2;

                                //S1 overlaps:
  if(((endm1s1 <= endm2s1) && (endm1s1 >= begm2s1)) ||
     ((begm1s1 >= begm2s1) && (begm1s1 <= endm2s1)))   //If they overlap,
    count++;                                           //mark it so.
                                //S2 overlaps:
  if(((endm1s2 <= endm2s2) && (endm1s2 >= begm2s2)) ||
     ((begm1s2 >= begm2s2) && (begm1s2 <= endm2s2)))
    count++;

  return count;
  }

int matchGetLength(match m)  {return m->length;}
int matchGetIndexS1(match m) {return m->indexS1;}
int matchGetIndexS2(match m) {return m->indexS2;}
int matchGetIsforward(match m) {return m->isforward;}
void displayMatch(match m) 
{
  fprintf(stderr, "Match: ");
  fprintf(stderr, "\tLength = %i, IndexS1 = %i, IndexS2 = %i, Overlaps = %i\n",
                  m->length, m->indexS1, m->indexS2, m->overlaps);
  return;
}


/////////////////////////////////////////////////////////////////////
//definitions for the matching stuff


/*Compare two Perfect Sequence Structures' lengths.
*/
int compByLen(const void* s1, const void* s2)
  {
  if((*(int**)s1)[0] < (*(int**)s2)[0])
    return -1;
  if((*(int**)s1)[0] > (*(int**)s2)[0])
    return 1;

  return 0;
  }



/*Repeatedly find the longest matching section from -s1- in -s2-.
  ** -lenS1- must be <= -lenS2- !
*/
match* matchingLoop(int* os1, int lenS1, int* os2, int lenS2, int* matchLen)
  {
  match m;
  int *s1, *s2;
  match* ret;
  int i;

  ret = (match*) calloc(lenS1, sizeof(match));
  *matchLen = 0;

  //copy s1 and s2 so we can mangle them with impunity
  s1 = (int*) calloc(lenS1, sizeof(int));
  s2 = (int*) calloc(lenS2, sizeof(int));
  for(i = 0; i < lenS1; ++i)
    s1[i] = os1[i];
  for(i = 0; i < lenS2; ++i)
    s2[i] = os2[i];



  theMatch* mlist; //List of longest matches found for use inside pickNextMatch.
  mlistsize = lenS1*lenS2 + 2; //The size of the list.
  //int mlistsize = 100*lenS1; //FIX THIS!!! this for when we run out of memory.
                             // (fix assert when changing this)
                             // (declared here for efficiency reasons)
  if(!(mlist = (theMatch*) calloc(mlistsize, sizeof(theMatch))))
    {
    char* errstr = strerror(errno);
    cerr << "ERROR: allocating " << mlistsize << " bytes: " << errstr << ".\n";
    exit(0);
    }

  m = pickNextMatch(s1, lenS1, s2, lenS2, mlist);
  while(m != 0)
    {
    ret[*matchLen] = m;            //Put the match into the ret array.
    ++(*matchLen);
                                   //Find the next match.
    m = pickNextMatch(s1, lenS1, s2, lenS2, mlist);
    }

  free(mlist);

  //printf("MatchLen = %i\n", *matchLen);
  //for(int i = 0; i < *matchLen; i++)
    //displayMatch(ret[i]);

  //fprintf(stderr, "renameMulti free3 %x %x\n", s1, s2);
  free(s1);
  free(s2);

  return ret;
  }


/* Pick the match of maximum length.  Or the best in case of ties.
Return the match or 0 if there are no matches to make.

  **SPEEDUPS: could start with longest matches and then quit after overlap
              is shorter than longest match.
*/
match pickNextMatch(int* s1, int lenS1, int* s2, int lenS2, theMatch* mlist)
  {
  int i, j, oc, bothCursor, oneCursor;
  int* rs1;                       //The reverse of s1;
  match m;
  theMatch* mBestTies;
  int numties = 0;
  theMatch* mMatchBothLR;
  theMatch* mMatchOneLR;
  int mlen = 0;                   //The current number of values in mlist.
  int lowind;                     //The index of a match with fewest overlaps.

  if(PICK_OVERLAPS + PICK_OVER_NEIGHBORS + PICK_RANDOM != 1)
    {                             //Check the PICK constants.
    fprintf(stderr, "PICK constants set improperly; only one should be 1.\n");
    exit(0);
    }

  rs1 = (int*) calloc(lenS1, sizeof(int));
  for(i=0; i < lenS1; i++)                //Get the reverse of s1:
    rs1[i] = -s1[lenS1-i-1];


                                          //Find the longest matches:
  findLongestMatches(s1, lenS1, s2, lenS2, mlist, &mlen, true);
  findLongestMatches(rs1, lenS1, s2, lenS2, mlist, &mlen, false);


  for(i=0; i < mlen; i++)                 //Innitialize the overlap values.
    mlist[i].overlaps = 0;
  for(i=0; i < mlen; i++)                 //Count the number of overlaps:
    for(j=i+1; j < mlen; j++)
      {                                   //Add the number of overlaps up
      oc = matchOverlapCount(mlist[i], mlist[j]);
      mlist[i].overlaps += oc;            //for each.
      mlist[j].overlaps += oc;
      }
  lowind = 0;
  for(i=0; i < mlen; i++)                 //Find a match with least overlaps.
    if(mlist[i].overlaps < mlist[lowind].overlaps)
      lowind = i;

//for(i=0; i < mlen; i++)                 //Check for matches of same length.
  //if(mlist[i].overlaps == mlist[lowind].overlaps && (lowind != i))
    //cerr << "DUP SHORTEST=-==-===-=-=-=-=-=-=-=-=-=-=" << endl;

  if(!mlen)                               //If there's no match, we're done:
    {
    free(rs1);
    return 0;
    }

  if(PICK_OVERLAPS || PICK_OVER_NEIGHBORS)
    {
    mBestTies = (theMatch*) calloc(mlen, sizeof(theMatch));
    for(i=0; i < mlen; i++)             //Get matches with least overlap.
      {                                   //**efficiency: build list first time.
      if(mlist[i].overlaps == mlist[lowind].overlaps)
        {
        mBestTies[numties] = mlist[i];
        ++numties;
        }
      }
      
    if(PICK_OVERLAPS)
      {
      oc = rand() % numties;
      m = copyMatch(&mBestTies[oc]);
      }
    else //PICK_OVER_NEIGHBORS
      {             //matches on left and right
      mMatchBothLR = (theMatch*) calloc(mlen, sizeof(theMatch));
                    //only matches on the left or the right
      mMatchOneLR = (theMatch*) calloc(mlen, sizeof(theMatch));
      bothCursor = 0;
      oneCursor = 0;
                    //get the side matches
      for(i = 0; i < numties; ++i)
        {
        if(mLeft(s1, lenS1, s2, lenS2, &(mBestTies[i])) && 
           mRight(s1, lenS1, s2, lenS2, &(mBestTies[i])))
          {
          mMatchBothLR[bothCursor] = mBestTies[i];
          ++bothCursor;
          }
        else if(mLeft(s1, lenS1, s2, lenS2, &(mBestTies[i])))
          {
          mMatchOneLR[oneCursor] = mBestTies[i];
          ++oneCursor;
          }
        else if(mRight(s1, lenS1, s2, lenS2, &(mBestTies[i])))
          {
          mMatchOneLR[oneCursor] = mBestTies[i];
          ++oneCursor;
          }
        }

          //pick the match
      if(bothCursor)
        m = copyMatch(&mMatchBothLR[rand() % bothCursor]);
      else if(oneCursor)
        m = copyMatch(&mMatchOneLR[rand() % oneCursor]);
      else
        m = copyMatch(&mBestTies[rand() % numties]);

      free(mMatchBothLR);
      free(mMatchOneLR);
      }
    free(mBestTies);
    }
  else if(PICK_RANDOM)
    m = copyMatch(&mlist[rand() % mlen]);
  else
    {
    fprintf(stderr, "You ****** it up in renameMulti.c, pickNextMatch\n");
    exit(0);
    }
  
  for(i=0; i < m->length; i++)            //Rename the elements as used:
    {
    s1[m->indexS1+i] = REMAPSTART;
    s2[m->indexS2+i] = REMAPSTART+1;
    }

  
  free(rs1);
  return m;
  }

//given s1, len s1, s2, lens2 and a match see if the elem to the left of the
// match is the same in both seqs
int mLeft(int* s1, int lenS1, int* s2, int lenS2, match m)
{
    if(m->indexS1 == 0 && m->indexS2 == 0)
        return 1;
    else if(m->indexS1 == 0 || m->indexS2 == 0)
        return 0;
    else
        return s1[m->indexS1 - 1] == s2[m->indexS2 - 1];
}

//given s1, len s1, s2, lens2 and a match see if the elem to the right of the
// match is the same in both seqs
int mRight(int* s1, int lenS1, int* s2, int lenS2, match m)
{
    if(m->indexS1 + m->length == lenS1 && m->indexS2 + m->length == lenS2)
        return 1;
    else if(m->indexS1 + m->length == lenS1 || m->indexS2 + m->length == lenS2)
        return 0;
    else
        return s1[m->indexS1 + m->length] == s2[m->indexS2 + m->length];
}

/* Find the longest substrings in both s1 and s2.  Keep all matches of
a particular length in case of ties. Puts the match in -mlist-.
-isf- should be true if -s1- is forward oriented.
*/
void findLongestMatches(int* s1, int lenS1, int* s2, int lenS2, theMatch* mlist,
                        int* mlen, bool isf)
  {
  int i,i1,i2;
  int length;      //The length of the longest match.
  int t;
  int temp[2];     //The indeces of the longest match.

  temp[0] = temp[1] = 0;
  
  if(*mlen)                      //Innitialize -mlen-
    length = mlist[0].length;   //to the largest length found so far
  else
    length = 1;                  //or to 1 if we haven't found any.
   
                                 //Sweep s1 over s2:
  for(i=1; i < lenS1; i++)       //Move s1 until the beginning lines up with s2:
    {
    i1 = lenS1-i;
    i2 = 0;
    while(i1 < lenS1 && i2 < lenS2)        //Get overlaps with this alignment:
      {
      t = 0;
                                           //Find the length of a match:
      while((i1 < lenS1 && i2 < lenS2) && (s1[i1++] == s2[i2++]))
        {
        if(t == 0)
          {
          temp[0] = i1-1;
          temp[1] = i2-1;
          }
        t++;
        }

      if(t == length)                      //If we've found an equal match,
        {                                  //add it to the list.
        fillMatch(mlist, *mlen, temp[0], temp[1], t, isf, lenS1);
        (*mlen)++;
        }
      
      if(t > length)                       //If we've found a longest match,
        {
        *mlen = 1;                         //put it as the only element.
        fillMatch(mlist, 0, temp[0], temp[1], t, isf, lenS1);
        length = t;
        }
       
      }
    }





  i = 0;
  while((i+lenS1) <= lenS2)      //Move s1 until the right sides line up:
    {
    i2 = i;
    for(i1=0; i1 < lenS1;)                 //Get overlaps with this alignment:
      {
      t = 0;
                                           //Find the length of a match:
      while((i1 < lenS1) && (s1[i1++] == s2[i2++]))
        {
        if(t == 0)
          {
          temp[0] = i1-1;
          temp[1] = i2-1;
          }
        t++;
        }


      if(t == length)                      //If we've found an equal match,
        {                                  //add it to the list.
        fillMatch(mlist, *mlen, temp[0], temp[1], t, isf, lenS1);
        (*mlen)++;
        }
      
      if(t > length)                       //If we've found a longest match,
        {
        *mlen = 1;                         //put it as the only element.
        fillMatch(mlist, 0, temp[0], temp[1], t, isf, lenS1);
        length = t;
        }
       
      }
    i++;
    }




                                 //Move s1 right of s2:
  for(i=lenS2-lenS1+1; i < lenS2; i++)
    {
    i1 = 0;
    i2 = i;
    while(i1 < lenS1 && i2 < lenS2)        //Get overlaps with this alignment:
      {
      t = 0;
                                           //Find the length of a match:
      while((i1 < lenS1 && i2 < lenS2) && (s1[i1++] == s2[i2++]))
        {
        if(t == 0)
          {
          temp[0] = i1-1;
          temp[1] = i2-1;
          }
        t++;
        }


      if(t == length)                      //If we've found an equal match,
        {                                  //add it to the list.
        fillMatch(mlist, *mlen, temp[0], temp[1], t, isf, lenS1);
        (*mlen)++;
        }
      
      if(t > length)                       //If we've found a longest match,
        {
        *mlen = 1;                         //put it as the only element.
        fillMatch(mlist, 0, temp[0], temp[1], t, isf, lenS1);
        length = t;
        }
       
      }
    }
  }



//Returns an integer which is an index of a longest candidate sequence for the
//matching, -1 if there isn't one.  The 0th entry is the length of the array.
//Takes the perfect seq we want to match with, the list of perfect seqs that we
//can match to, and the length of the possible matches array.
int getLongestMatch(int* pseq, int** possMatches, int lenPossMatches,
                    int& size, bool& overlaps)
  {
  int index = -1;
  size = 0;

  /*
  printf("matching ");
  displaySingSeq(pseq, 0);
  printf("\n");

  printf("choices \n");
  displaySeqs(possMatches, lenPossMatches);
  */

  for(int i = 0; i < lenPossMatches; ++i)
    if(doesOverlap(pseq, possMatches[i]) || !overlaps)//If the sequence overlaps
      {                                               //or no previous has yet:
      overlaps = doesOverlap(pseq, possMatches[i]);
      int ol = overlapSize(pseq, possMatches[i]);

      if(size < ol)        //If we've found a larget overlap
        {
        size = ol;         //then mak it as such.
        index = i;
        }
      }

  return index;
  }



//returns an array of integers, each one is an index of a candiadate sequence
//for the matching, the 0th entry is the length of the array
//takes the perfect seq we want to matchwith, the list of perfectseqs that we
//can match to and the length of the possible matches array
int* getPossMatches(int* pseq, int** possMatches, int lenPossMatches)
  {
  int* res;
  int i, matchCount;

  matchCount = 0;
  res = (int*) calloc(lenPossMatches + 1, sizeof(int));

  /*
  printf("matching ");
  displaySingSeq(pseq, 0);
  printf("\n");

  printf("choices \n");
  displaySeqs(possMatches, lenPossMatches);
  */

  for(i = 0; i < lenPossMatches; ++i)
    {
    if(overlapSize(pseq, possMatches[i]))
      {
      ++matchCount;
      res[matchCount] = i;
      }
    }

  res[0] = matchCount;

  return res;
  }

//=============================
//Determine the size of the overlaping regions in the perfect sequences.
int overlapSize(int* ps1, int* ps2)
  {
  int i, j, intersectSize;
  intersectSize = 0;

  for(i = 1; i <= psLength(ps1); ++i)
    {
    for(j = 1; j <= psLength(ps2); ++j)
      {
      if(ps1[i] == ps2[j])
        {
        //funky but ok, draw the pictures for the 4 cases
        if(ps1[0] - i < ps2[0] - j)
          intersectSize = ps1[0] - i + 1;
        else
          intersectSize = ps2[0] - j + 1; 
        j = psLength(ps2) + 1; //break out of the loop
        i = psLength(ps1) + 1;
        }
      }
    }

  return intersectSize;
  }


//the start of the overlap region in ps1, there must be an overlap
int overlapLocS1(int* ps1, int* ps2)
{
  int i, j;

  for(i = 1; i <= psLength(ps1); ++i)
    {
      for(j = 1; j <= psLength(ps2); ++j)
        {
         if(ps1[i] == ps2[j])
            return i;
        }
    }
  return 0;
}

//the start of the overlap region in ps1, there must be an overlap
int overlapLocS2(int* ps1, int* ps2)
{
  int i, j;

  for(i = 1; i <= psLength(ps1); ++i)
    {
      for(j = 1; j <= psLength(ps2); ++j)
        {
          if(ps1[i] == ps2[j])
             return j;
        }
    }
  return 0;
}

//we set all the elements in the match in s1 and s2 to -1 so they don't get matched again
/*
void renameSeqs(int* s1, int* s2, match m)
{
  int i;
  
  for(i = 0; i < psLength(matchGetPerfSeq(m)); ++i)
    {
      s1[matchGetIndexS1(m) + i] = -1;
      s2[matchGetIndexS2(m) + i] = -1;
    }
  return;
}
*/

////////////////////////////////////////////////////////////////////
//definitions for the perfect sequence stuff

//this extracts the perfect sequences on input which has 0 based length of
// seqLen and return a list of the perfect sequences as res
//the number of perfect sequences is in numSeq. Further this ignores any
// sequences of -1's
int** extractPerfSeq(int* input, int seqLen, int* numSeq)
  {
  int** res;
  int* seqPtr;
  int startLoc;
  int endLoc;
  int i;

  res = (int**) calloc(seqLen + 1, sizeof(int*));
  *numSeq = 0;
  startLoc = 0;
  endLoc = 1;

  while(startLoc < seqLen)
    {
    //Scan out the already visited elements:
    while(startLoc < seqLen && input[startLoc] == REMAPSTART)
      ++startLoc;

    endLoc = startLoc + 1;

    //if we are basically off the end of the sequence break out of the loop
    if(startLoc == seqLen)
      break;

    //if the next elem is part of the perfectly ordered seq
    while(endLoc < seqLen && input[endLoc] == input[endLoc - 1] + 1)
      ++endLoc;

    //so now startLoc is the first elem in a perf order seq and endLoc - 1 is
    // the last elem in the seq
    //thus there are (endLoc - startLoc) elems in the seq, plus one for the
    // length info, 1 for the startIndex and 1 cause I am scared
    seqPtr = (int*) calloc((endLoc - startLoc) + 3, sizeof(int));
    seqPtr[0] = endLoc - startLoc;
    seqPtr[seqPtr[0] + 1] = startLoc;
    absCopy((input + startLoc), (input + endLoc), seqPtr);
    res[*numSeq] = seqPtr;
    ++(*numSeq);

    startLoc = endLoc;
    }

  for(i = 0; i < *numSeq; ++i)
    {
    if(res[i][1] < 0)
      normalizeSegment(res[i]);
    }
  
  return res;
  }

int* copyPs(int* ps)
{
  int* newPS;
  int i;

  newPS = (int* ) calloc(ps[0] + 3, sizeof(int));

  for(i = 0; i <= ps[0] + 1; ++i)
    newPS[i] = ps[i];

  return newPS;
}

int psLength(int* ps) {return ps[0];}
int psStartInd(int* ps) {return ps[ps[0] + 1];}
int psOrient(int* ps, int* psSourceSeq)
  {return extractSign(psSourceSeq[psStartInd(ps)]);}

void freePerfSeqs(int** ps, int lenPS)
{
  int i;
  for(i = 0; i < lenPS; ++i)
    free(ps[i]);

  //fprintf(stderr, "renameMulti free5 %x\n", ps);
  free(ps);
  return;
}


int extractSign(int x)
{
  if(x >= 0)
    return 1;
  else
    return -1;
}

/* Return true iff -s1- overlaps(non-containment) either side of the interval 
defined by -s2-.
*/
bool doesOverlap(int* s1, int* s2)
  {
  if((s1[1] <= s2[1] && s1[psLength(s1)-1] >= s2[1]) ||
     (s1[1] <= s2[psLength(s2)-1] && s1[psLength(s1)-1] >= s2[psLength(s2)-1]))
    return true;

  return false;
  }

//////////////////////////////////////////////////////////////
//The random stuff for dealing with sequences etc.

/*Rename both sequences so that -s1- is the identity.
  -s1- cannot contain copies!
*/
void makeIdentity(int* s1, int ls1, int* s2, int ls2)
  {
  int i;
  hash_map< int, int, hash<int> > remap;        //Holds the remap values.

  int mark=1;
  for(i=0; i < ls1; i++)                        //Remap s1:
    {
    if(s1[i] < 0)
      remap[abs(s1[i])] = -mark;
    else
      remap[abs(s1[i])] = mark;

    s1[i] = mark++;
    }

  int highest = ls1+1;
  for(i=0; i < ls2; i++)                        //Remap s2: 
    {
    if(remap[abs(s2[i])])
      {
      if(s2[i] < 0)
        s2[i] = -remap[abs(s2[i])];
      else
        s2[i] = remap[abs(s2[i])];
      }
    else                                        //If it wasn't seen in s1,
      {
      if(s2[i] < 0)                             //make sure it's higher than
        s2[i] = -(highest++);                   //the highes s1.
      else
        s2[i] = highest++;
      }
    }
  }

/*
void updateSeqMatch(int* s1, int* s2, match m)
{
  int i;

  //clear the matched stuff in s1
  for(i = 0; i < psLength(matchGetPerfSeq(m)); ++i)
    s1[matchGetIndexS1(m) + i] = REMAPSTART;

  //clear the matched stuff in s2
  for(i = 0; i < psLength(matchGetPerfSeq(m)); ++i)
    s2[matchGetIndexS2(m) + i] = REMAPSTART;

  return;
}
*/


/* Return the value of the maximum element from both sequences.
*/
int getMax(int* s1, int* s2, int ls1, int ls2)
  {
  int max = 0;
  int i;

  for(i=0; i < ls1; i++)
    if(abs(s1[i]) > max)
      max = abs(s1[i]);

  for(i=0; i < ls2; i++)
    if(abs(s2[i]) > max)
      max = abs(s2[i]);

  return max;
  }

/* Rename all the corresponding elements in the sequence.
    - DOES NOT preserve contiguous/non-contiguous relationships among the 
      duplicates that don't have a match.
*/
void doCopyRenaming(int* s1, int* s2, int ls1, int ls2, match* allMatches,
                    int allMatchesLen)
  {
  int i, j, remap;
  match m;
  bool didremap, *remaps1, *remaps2;
  remaps1 = (bool *)malloc(ls1 * sizeof(bool));
  remaps2 = (bool *)malloc(ls2 * sizeof(bool));

  for(i=0; i < ls1; i++)    //Initialize remaps arrays:
    remaps1[i] = true;
  for(i=0; i < ls2; i++)
    remaps2[i] = true;

                            //Rename all matches (other than the first
                            //one we see). Mark all those that aren't in a match
                            //to be renamed:
  remap = REMAPSTART;
  for(i = 0; i < allMatchesLen; ++i)        //For each match in allMatches,
    {
    m = allMatches[i];

    int length = matchGetLength(m);
    didremap = false;                       //True if we had to remap anything.
    for(j = 0; j < length; j++)             //Rename the substring:
      {
      int index = matchGetIndexS1(m);
                                            //Rename in s1...
      if(matchGetIsforward(m))              //Check for rename orientation.
        {
        if(s1[index+j] > 0)                 //Check the sign of the renamed.
          s1[index+j] = remap+j;
        else
          s1[index+j] = -(remap+j);
        }
      else
        {
        if(s1[index + (length-1-j)] > 0)
          s1[index + (length-1-j)] = remap+j;
        else
          s1[index + (length-1-j)] = -(remap+j);
        }
                                            //Rename in s2...
      if(s2[matchGetIndexS2(m)+j] > 0)      //Check the sign of the renamed.
        s2[matchGetIndexS2(m)+j] = remap+j;
      else
        s2[matchGetIndexS2(m)+j] = -(remap+j);

      didremap = true;                          //Mark that we did a remap.

      remaps1[matchGetIndexS1(m) + j] = false;  //Mark as not to be remapped.
      remaps2[matchGetIndexS2(m) + j] = false;  //Mark as not to be remapped.
      }

    if(didremap)                                //If we did a remap,
      remap += length+1;                        //increment the remap.
    }

  
  /*
  cerr << "rem:\t";
  for(int j=0; j<ls1; j++)
    cerr << remaps1[j] << " ";
  cerr << "\n\t";
  for(int j=0; j<ls2; j++)
    cerr << remaps2[j] << " ";
  cerr << "\n";
  */

  remap = relabelExtras(s1, ls1, remaps1, remap);
  relabelExtras(s2, ls2, remaps2, remap);


  free(remaps1);
  free(remaps2);

  return;
  }




/* Rename the copies in a single sequence.  These copies are those that will
be considered insertions or deletions so it doesn't matter what we rename them
to.
*/
void renameSingleSideCopies(int* s, int ls)
  {
  int maxel = getMax(s, (int*)NULL, ls, 0);             //The largest in s.
  bool* beenseen = (bool*)calloc(maxel, sizeof(bool*)); //Mark values seen.
  int next = maxel+1;
  int i;

  for(i=0; i < maxel; i++)
    beenseen[i] = false;

  for(i=0; i < ls; i++)
    {
    if(beenseen[abs(s[i])])
      s[i] = next++;
    else
      beenseen[abs(s[i])] = true;
    }

  free(beenseen);
  }




/* Rename all sequence values that are marked by true in -rena- (those values
that are insertions/deletions).  Return a label value higher than any used.
The contiguous/non-contiguous relationships among the elements to be renamed 
will NOT be maintained.
    - Start the naming at -start-.
*/
int relabelExtras(int* seq, int slen, bool* rena, int start)
  {
  int i;

  for(i=0; i < slen; i++)
    if(rena[i])                          //If this element in -seqs- will be
      seq[i] = start++;                  //renamed, then do so.

  return start+1;
  }


/* Rename both sequences so that one of the two contain each element from 1 to
max(-s1-, -s2-).
*/
void collapseGaps(int* s1, int ls1, int* s2, int ls2)
  {
  int i;
  int max = 0;                              //The maximum value in both.
  hash_map< int, int, hash<int> > beenseen; //True if the value has been seen.
  hash_map< int, int, hash<int> > remap;    //The value we will remap to.


  for(i=0; i < ls1; i++)                    //Init beenseen and max:
    {
    beenseen[abs(s1[i])] = 1;
    if(max < abs(s1[i]))
      max = abs(s1[i]);
    }
  for(i=0; i < ls2; i++)
    {
    beenseen[abs(s2[i])] = 1;
    if(max < abs(s2[i]))
      max = abs(s2[i]);
    }

  int current = 1;
  for(i=1; i <= max; i++)
    if(beenseen[i])
      remap[i] = current++;

  for(i=0; i < ls1; i++)                    //Rename the sequences:
    {
    if(s1[i] < 0)
      s1[i] = -remap[abs(s1[i])];
    else
      s1[i] = remap[abs(s1[i])];
    }
  for(i=0; i < ls2; i++)
    {
    if(s2[i] < 0)
      s2[i] = -remap[abs(s2[i])];
    else
      s2[i] = remap[abs(s2[i])];
    }
  }


/* Reverse all negated sections of the given string.
*/
void normalizeSeq(int* seq, int seqlen)
  {
  int right,left,j, i=0;
  int temp;
  while(i < seqlen)
    {
    if(seq[i] < 0)                             //If a negative is found
      {
      j = i;
      left = i;
 
      while((j < seqlen) && (seq[++j] < 0));   //find the first positive,
      right = j-1;
      while(left <= right)                     //and swap all the negatives.
        {
        temp = abs(seq[left]);
        seq[left] = abs(seq[right]);
        seq[right] = temp;
        left++;
        right--;
        }
      i = j;
      }
    else
      i++;
    }
  }


/* Output a sequence to stdout.
*/
void printSeq(const char* title, int* seq, int len)
  {
  fprintf(stderr, "%s:\n", title);
  for(int i = 0; i < len; ++i)
    fprintf(stderr, "[%i]%i\t", i, seq[i]);
  fprintf(stderr, "\n");
  }
