//Krister Swenson, Mark Marron                                    FALL 2003
#ifndef RENAMEMULTI
#define RENAMEMULTI

#ifdef _MSC_VER
#pragma warning (disable : 4786) // Turn off dumb Visual C++ warning
#endif

#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <ctime>
#include <vector>
#include <ext/hash_map>
#include <errno.h>
#include <assert.h>
#include <string.h>
#include "../graph/Pair.h"
#include "../graph/gencov.h"
#include "DupInfo.h"



//Set one of the 3 to 1 (the others to 0) and that tie-breaker will be used.
//random tie-breaker:
#define PICK_RANDOM 0
//tie-breaker based on number of overlaps:
#define PICK_OVERLAPS 0
//tie-breaker based on overlaps and then similar neighbors:
#define PICK_OVER_NEIGHBORS 1


   //FIX:Just set this when we get the sequences!
#define REMAPSTART 90001 //The value to start renaming at. It must be greater
                         // than the largest value in both sequences.


//define what a matching is
//we use a structure to keep track of several things about a match
//(1) the length of the matching section
//(2) The index of the first element in s1.
//(3) The index of the first element in s2.
//(4) isforward: true if s2 matched the reverse of something in s1.
//(5) overlaps: true if the the match doesn't overlap another match of the
//              same length.

typedef struct mat
{
  int length;
  int indexS1;
  int indexS2;
  bool isforward;
  int overlaps;
} theMatch;
typedef struct mat* match;



//THE INTERFACE TO ALL THESE FUNCTIONS:
  // Pick the original for each copy existing in both sequences.  Rename the
  // them accordingly.
void renameMulti(int* s1, int ls1, int* s2, int ls2);

  //Pair up the duplicates. Return an array -A- where A[i] is the assignment
  //such that s1[i] = A[i].  A[i] < 0 if there is no corresponding gene.
  //    ** user resposible for the array memory returned.
int* getDupArray(int* s1, int ls1, int* s2, int ls2);

  //Pair up the duplicates. Return a vector of structures relating the elements.
std::vector<DupInfo> getDupInfo(int* s1, int ls1, int* s2, int ls2);

  //Remove all genes that have duplicates from both sequences.
void removeDups(int* s1, int &ls1, int* s2, int &ls2);




//define the functions for manipulating matches
void fillMatch(match* mlist, int ind,  int i1, int i2, int len,
               bool isf, int S1len);
match copyMatch(match mat);
match makeMatch(int len, int oS1, int iS1, int oS2, int iS2);
int matchOverlapCount(match m1, match m2);
void freeMatch(match m);
int matchGetLength(match m);
int matchGetIndexS1(match m);
int matchGetIndexS2(match m);
int matchGetIsforward(match m);
void displayMatch(match m);

//The stuff to do the matching:
  //Repeatedly find the longest matching section from -s1- in -s2-.
  //  ** -lenS1- must be <= -lenS2-!
match* matchingLoop(int* s1, int lenS1, int* s2, int lenS2, int* matchLen); 
  //Pick the match of maximum length.
  //Return the match or 0 if there are no matches to make.
match pickNextMatch(int* s1, int lenS1, int* s2, int lenS2, theMatch* mlist);

//given s1, len s1, s2, lens2 and a match see if the elem to the left of 
//the match is the same in both seqs.
int mLeft(int* s1, int lenS1, int* s2, int lenS2, match m);

//given s1, len s1, s2, lens2 and a match see if the elem to the right of 
//the match is the same in both seqs.
int mRight(int* s1, int lenS1, int* s2, int lenS2, match m);

  //Find the longest substrings in both s1 and s2.  Keep all matches of
  //a particular length in case of ties. Puts the match in -mlist-.
  //-isf- should be true if -s1- is forward oriented.
void findLongestMatches(int* s1, int lenS1, int* s2, int lenS2, theMatch* mlist,
                        int* mlen, bool isf);
  //determine the size of the overlaping regions in the perfect sequences
int overlapSize(int* ps1, int* ps2); 
  //the start of the overlap region in ps1, there must be an overlap
int overlapLocS1(int* ps1, int* ps2); 
  //the start of the overlap region in ps1, there must be an overlap
int overlapLocS2(int* ps1, int* ps2); 
  //rename the sequences to ack the match m
void renameSeqs(int* s1, int* s2, match m); 
  //Rename the copies in a single sequence.  These copies are those that will
  //be considered insertions or deletions so it doesn't matter what we rename
  //them to.
void renameSingleSideCopies(int* s, int ls);
int* getPossMatches(int* pseq, int** possMatches, int lenPossMatches);
int getLongestMatch(int* pseq, int** possMatches, int lenPossMatches,
                    int& size, bool& abudts);


//The Stuff to deal with the perfect(contiguous) Sequences
//NOTE on the hacked perfSeqs, ps[0] is the length of the perfSeq,
// index ps[1] is the first index that this seq appears at in s.
int** extractPerfSeq(int* input, int seqLen, int* numSeq);
int* copyPs(int* ps);
int psLength(int* ps);
int psStartInd(int* ps);
int psOrient(int* ps, int* psSourceSeq);
void freePerfSeqs(int** ps, int lenPS);
int extractSign(int x);
bool doesOverlap(int* s1, int* s2);
int compByLen(const void* s1, const void* s2);

//The random stuff for dealing with sequences etc.
  //update s1 and s2 to reflect the match made by m.
void updateSeqMatch(int* s1, int* s2, match m);
  //rename s1 and s2 according to the matchings.
void doCopyRenaming(int* s1, int* s2, int ls1, int ls2, match* allMatches,
                    int allMatchesLen);
int getMax(int* s1, int* s2, int ls1, int ls2);
  //Rename the insertions and deletions.
  //Return a label value higher than any used.
int relabelExtras(int* seq, int slen, bool* rename, int start);
  //Rename both sequences so that -s1- is the identity.
  //-s1- cannot contain copies!
void makeIdentity(int* s1, int ls1, int* s2, int ls2);
  //Rename both sequences so that one of the two contain each element from 1 to
  //max(-s1-, -s2-).
void collapseGaps(int* s1, int ls1, int* s2, int ls2);
  //Reverse all negated sections of the given string.
void normalizeSeq(int* seq, int seqlen);

  //Output a sequence.
void printSeq(const char* title, int* seq, int len);




#endif
