/* Krister Swenson                                                 Winter 2007-8

Excercise the mutual noninterfering computations.
*/

#include <iostream>
#include <vector>
#include <algorithm>
#include "../graph/Ograph.h"

#define REPS 1000
#define PERMLEN 12

using namespace std; 



bool isCommuting(const list<Pair> &set)
  {
  for(list<Pair>::const_iterator i = set.begin(); i != set.end(); i++)
    {
    list<Pair>::const_iterator j = i;
    while(++j != set.end())
      if(i->overlaps(*j))     //If the two intervals overlap then sort the 
        {                     //intervals and print message:
        vector<Pair> setv;
        std::copy(set.begin (), set.end (), back_inserter(setv));
        sort(setv.begin(), setv.end());
        cerr << "sorted NMIS:" << setv << endl;
        cerr << "offending pairs:" << *i << *j << endl;

        return false;
        }
    }

  return true;
  }


int cgsseen=1;      //Count the number of non-circle graphs encountered.
int counter=1;      //Count the number of trials so far.
int trials=1;       //Count the total number of trials so far.
float total=0;      //The total of all |MIS|/d() ratios so far.
void tryPair(const vector<int> &perm1, const Ograph &g1,
             const vector<int> &perm2, const Ograph &g2)
  {
  int dist1 = g1.getDistNH();
  int dist2 = g2.getDistNH();

  cerr << "--------------" << endl;
  cerr << "trying[" << counter++ << "-" << cgsseen << "]:\n\t"
       << perm1 << "\n\t" << perm2 << endl;

  list<Pair> nmis = g1.getMutualNonintMIS(g2);
  cerr << "\tMutual NMIS:" << nmis << endl;

                                 //Sanity check:
  assert(g1.isNonint(nmis));          //Make sure they sort (and commute).
  assert(g2.isNonint(nmis));          //Make sure they sort (and commute).
  assert(isCommuting(nmis));          //Double check they commute.

  if(dist1 > 0)                   //Compute size ratios:
    {
    float ratio1 = (float)nmis.size()/(float)dist1;
    total += ratio1;
    cerr << "\tratio1:" << nmis.size() << "/" << dist1 << "=" << ratio1 << endl;
    }
  if(dist2 > 0)                   //Compute size ratios:
    {
    float ratio2 = (float)nmis.size()/(float)dist2;
    total += ratio2;
    float avgratio = total/(float)trials++;
    cerr << "\tratio2:" << nmis.size() << "/" << dist2 << "=" << ratio2
         << "  average:" << avgratio << endl;
    }
  }

void checkCG(const vector<int> &perm, const Ograph &g)
  {
  if(!g.isCG())                      //test whether it has a chord model.
    {
    cerr << "Not a circle graph!: " << perm << endl;
    //int sum=0, psum=0;
    //for(unsigned int i=0; i<perm.size(); i++)
      //{ sum += i+1; psum += perm[i]; }
    //if(sum != psum)                  //If it's not a positive permutation.
    cgsseen++;
    //exit(1);
    }
  }


/*Make -perm- a random permutation and return the corresponding HP graph.
*/
Ograph randomizePerm(vector<int> &perm1)
  {
  int tmp[3] = {1,2,3};
  Ograph g(tmp,3);
  Ograph h(tmp,3);
  vector<Pair> dummy;
  do
    {
    dummy.clear();
    for(unsigned int j=0; j < perm1.size(); j++)
      {                                //Randomize the permutation:
      int k = random()%perm1.size();    //Swap perm[j] with a random element.
      int temp = perm1[k];
      perm1[k] = perm1[j];
      perm1[j] = temp;
      }
    for(unsigned int j=0; j < perm1.size(); j++)
      if(random()%2)                   //Assign signs randomly:
        perm1[j] *= -1;

    g = Ograph(perm1);                  //Create the HP graph and
    h = g;                             //Do this because resolve() mutates h.
    }                                  //Go until we have no hurdles.
  while(h.getDistNH() != h.resolve(dummy));

  return g;
  }





int main()
  {
  vector<int> perm1, perm2;

  for(int i=1; i <= PERMLEN; i++)        //Create the identity:
    {
    perm1.push_back(i);
    perm2.push_back(i);
    }

  cerr << "Permutation Length:" << perm1.size() << endl;
  for(int i=0; i < REPS; i++)
    {
    Ograph g1 = randomizePerm(perm1);
    Ograph g2 = randomizePerm(perm2);

    tryPair(perm1, g1, perm2, g2);
    }


  cerr << "done." << endl;
  return(0);
  }
