/* Krister Swenson                                                 Winter 2007-8

Sort random permutations by maximum sets of noninterfering inversions.
*/

#include <math.h>
#include <iostream>
#include <vector>
#include <algorithm>
#include <numeric>
#include "../graph/Ograph.h"
#include "../tools/tools.h"

#define REPS 10
#define PERMLEN 8

using namespace std; 



bool isCommuting(const list<Pair> &set)
  {
  for(list<Pair>::const_iterator i = set.begin(); i != set.end(); i++)
    {
    list<Pair>::const_iterator j = i;
    while(++j != set.end())
      if(i->overlaps(*j))     //If the two intervals overlap then sort the 
        {                     //intervals and print message:
        vector<Pair> setv;
        std::copy(set.begin (), set.end (), back_inserter(setv));
        sort(setv.begin(), setv.end());
        cerr << "sorted NMIS:" << setv << endl;
        cerr << "offending pairs:" << *i << *j << endl;

        return false;
        }
    }

  return true;
  }



/* Run a test for a single permutation.
*/
int cgsseen=1;      //Count the number of non-circle graphs encountered.
int counter=1;      //Count the number of trials so far.
int trials=1;       //Count the total number of trials so far.
float total=0;      //The total of all |MIS|/d() ratios so far.

void tryPerm(const vector<int> &perm1, Ograph g1, vector<int> &depth)
  {
  cerr << "--------------" << endl;
  cerr << "trying[" << counter++ << "-" << cgsseen << "]:\n\t" << perm1 << endl;

  int numreps = 0;
  int dist = g1.getDistNH();
  while(dist)                              //While it's not sorted
    {
    numreps++;
    list<Pair> nmis = g1.getNonintMIS();   //find a maximum noninterfering set,
    cerr << "\tNMIS:" << nmis << endl;
                                           //Sanity check:  ___________________
    assert(g1.isNonint(nmis));             // Make sure they sort (and commute).
    assert(isCommuting(nmis));             // Double check they commute.
  
    if(dist > 0)                   //Compute size ratios:
      {
      float ratio1 = (float)nmis.size()/(float)dist;
      total += ratio1;
      float avgratio = total/(float)trials++;
      cerr << "\tratio:" << nmis.size() << "/" << dist << "=" << ratio1
           << "  average:" << avgratio << endl;
      }

    g1.applyCommutingSet(nmis);            //then apply the set,
    dist = g1.getDistNH();                 //and check the new distance.
    }

  cerr << "depth:" << numreps << endl;
  depth.push_back(numreps);
  }



/*Make -perm- a random permutation and return the corresponding HP graph.
*/
Ograph randomizePerm(vector<int> &perm1)
  {
  int tmp[3] = {1,2,3};
  Ograph g(tmp,3);
  Ograph h(tmp,3);
  vector<Pair> dummy;
  do
    {
    dummy.clear();
    for(unsigned int j=0; j < perm1.size(); j++)
      {                                //Randomize the permutation:
      int k = random()%perm1.size();    //Swap perm[j] with a random element.
      int temp = perm1[k];
      perm1[k] = perm1[j];
      perm1[j] = temp;
      }
    for(unsigned int j=0; j < perm1.size(); j++)
      if(random()%2)                   //Assign signs randomly:
        perm1[j] *= -1;

    g = Ograph(perm1);                  //Create the HP graph and
    h = g;                             //Do this because resolve() mutates h.
    }                                  //Go until we have no hurdles.
  while(h.getDistNH() != h.resolve(dummy));

  return g;
  }


int main()
  {
  vector<int> perm1;

  for(int i=1; i <= PERMLEN; i++)        //Create the identity:
    perm1.push_back(i);

  cerr << "Permutation Length:" << perm1.size() << endl;
  vector<int> depth;     //Vector of depths (iterations of NINT applications).
  vector<int> dists;     //Vector of distances.
  for(int i=0; i < REPS; i++)
    {
    Ograph g1 = randomizePerm(perm1);
    dists.push_back(g1.getDistNH());

    tryPerm(perm1, g1, depth);
    }

  cerr << "mean:" << getMean(depth) << endl
       << "deviation:" << getStandardDev(depth) << endl;

  cerr << "done." << endl;
  return(0);
  }
