/* Krister Swenson                                                  Spring 2006

Provide functions for dealing with sorting lattices.
*/
#ifndef LATTICEFUNCS_H
#define LATTICEFUNCS_H

#include <iostream>
#include <vector>
#include <string>
#include <queue>
#include <ext/hash_map>
#include "seqfuncs.h"
#include "../graph/med.h"
#include "../tools/Vertex.h"



  //Find an island of size equal to the inversion distance. If one exists return
  // true.
bool hasIsland(Ograph &gr);

  //Find an island of size equal to the inversion distance. If one exists return
  // true.
  // We do this by building a connected component (not including links
  // through the id) from an upper bound's child.  If after this is done there
  // are any unvisited children we know we have an island.
  //    ** leaves everything unmarked.
bool hasIsland(Vertex &ub, Vertex &lb);

  //Mark all vertices reachable from -v-.
void markComponent(Vertex &v);

  //Return true if one vertex is reachable from the other.
bool isReachable(Vertex &v, const Vertex &target);

  //Unmark v's connected component induced by considering only marked vertices.
void clean(Vertex &v);

  //Get the sorting lattice.  Everything is innitialize unmarked.
void fullLattice(__gnu_cxx::hash_map<unsigned int, Vertex> &vset,
                 Vertex &v, Ograph &gr, const std::string &str);

  //Given the upper bound, traverse the graph and only print the downward edges
  // so that dot makes it look like a lattice.
  //    (i.e. do a breadth-first traversal)
  //  ** leaves things marked.
void printLattice(Vertex &ub);

  //Traverse the graph and print the edges.
  //  ** leaves things marked.
void printGraph(Vertex &v);


  //Lookup function for a hash_map.  Returns 1 if -i- is in -Map-.
bool lookup(const __gnu_cxx::hash_map<unsigned int,
            Vertex>& Map, const unsigned i);
  //Get a hash value for a string.
  //   -- uses the algorithm from "art of programming"
unsigned int DEKHash(const std::string& str);



#endif
