/* Krister Swenson                                                  Spring 2006

Sequence functions.
*/
#ifndef SEQFUNCS_H
#define SEQFUNCS_H


#include <iostream>
#include <stdio.h>
#include <vector>
#include <list>
#include <utility>
#include <set>
#include <string>
#include <sstream>
#include <stack>
#include <assert.h>

#define MAXSEQLEN 100000 //The maximum size any read string will be.

using namespace std;

typedef pair<int,int> ipair;


  //Return true if there is a hurdle in the given permutation.
bool hasHurdle(const vector<int> p);


  //Input: Permutation
  //Output: Indices of endpoints of each hurdle in the permutation or empty
  //        vector if there is no hurdle.
vector<ipair> getUnorientedComps(const vector<int> perm);


  //Return a count of all positive permutations of length -len- with len+1
  // breakpoints.
unsigned long getBPPosCount(int len);

  //Return a count of all permutations of length -len- with len+1 breakpoints.
  //  ** much less resource dependent than getBPPerm.
unsigned long getBPPermCount(int len);

  //Return a vector of all permutations of length -len- with only breakpoints.
vector< vector<int> > getBPPerm(int len);
void getBPPerm(vector< vector<int> > &ret, vector<int> seq, vector<int> L,
               int SEQLEN);


  //Return true if the given permutation has an adjacency.
bool hasAdjacency(const vector<int> p);

  //Return a vector version of the given permutation.
vector<int> getVector(int* perm, int len);

  //Convert a list of inversions in endpoint format to a list of inversions
  // in set format.
list< set<int> > intTOset(list< pair<int,int> > invs, vector<int> perm);

  //Return a random permutation. Do srandom((unsigned int)time(NULL)) first.
vector<int> getRandPerm(int len);
  //Return a random signed permutation. Do srandom((unsigned int)time(NULL)).
vector<int> getRandSPerm(int len);

  //Label seq2 as though we were renaming s1 to be the identity.
void seqACCseq(const int s1[], int s2[], int len);
void seqACCseq(const vector<int> s1, vector<int> &s2);

  //Read in an input sequence and return an integer array that is the sequence.
  // ** User responsable for the memory.
int getInputSeq(int** ret);
vector<int> getInputSeq();

  //Make sure that the given sequence is a valid permutation of the identity.
bool validPerm(const int seq[], const int len);

  //Return s1 renamed as though s2 were the target.
int* renameSeq(const int s1[], const int s1len,
               const int s2[], const int s2len);

  //Return the identity of the given length.
vector<int> makeIDv(const int len);
  //   ** User resposible for the memory.
int* makeID(const int len);

  //Apply the given reversal to the given permutation.
  //The reversals are 1 indexed and inclusive.
void applyRev(vector<int> &p, ipair rev);


  //Rename the given sequence to be the identity; return the identity renamed
  //accordingly.
void invertSeq(vector<int>);
void invertSeq(int seq[], int len);

  //Print a sequence.
void printSeq(const int s[], const int slen);

  //Return a copy the sequence.
  //  ** User resposible for the memory.
int* copySeq(const int* s, const int len);

  //Return a string of the given sequence.
string seqTOstring(int* seq, int seqsize);


  //Print a vector.
template<typename T>
void printVector(vector<T> &v)
  {
  for(int i=0; size_t(i)<v.size(); i++)
    cerr << v[i] << " ";
  }

  //Print a list.
/*
void printList(list<int> &l)
  {
  list<int>::iterator first = l.begin();
  list<int>::iterator last = l.end();

  cerr << *first++;
  while(first != last) 
     cerr << " " << *first++;
  }
template<typename S>
void printList(list<S> &l)
  {
  list<S>::iterator first = l.begin();
  list<S>::iterator last = l.end();

  cerr << *first++;
  while(first != last) 
     cerr << " " << *first++;
  }
*/


#endif
