/* Krister Swenson                                                  Spring 2006

Produce a quartet where all pairwise distances are equal.
*/

#include "../seqtools/seqfuncs.h"
#include "../graph/Ograph.h"
#include "../multi/multiMED.h"
#include "sigfuncs.h"
#include <vector>

using namespace std; 


Pair getRotatingInv(const Ograph &from, const Ograph &around);
Pair getRotatingInv(const Ograph &from1, const Ograph &from2,
                    const Ograph &around);

void method2(int len, int dist);
void method1(int len, int dist);


int main()
  {
  int len, dist;
  cout << "Enter the sequence length: ";
  cin >> len;
  cout << "Enter the distance (<" << len << "): ";
  cin >> dist;
  assert(dist < len);

  method2(len, dist);
  }


void method2(int len, int dist)
  {
  int* idstr = makeID(len);      //Build the identity sequence.
  Ograph s1(idstr, len);         //Build the identity for seq1.
  Ograph s2(idstr, len);         //Build the identity for seq2.

          //Make s1 be dist/4 from the ID:
  for(int i=0; i < dist/4;i++)
    {
    s1.revAll(1,1+i);
    s2.revAll(len-i,len);
    }

          //Set up the graphs for s2 and s1 renamed according to each:
  int* s1seq = s1.getSeq();
  int* s2seq = s2.getSeq();

  Ograph s1As1(idstr, len);      //s1 named according to seq1.

  Ograph s2lAs2(idstr, len);     //s2 named according to seq2 -- left side.
  Ograph s2rAs2(idstr, len);     //s2 named according to seq2 -- right side.
  seqACCseq(s1seq, s2seq, len);  //Get s2 named as though s1 were the ID.
  Ograph s2lAs1(s2seq, len);     //s2 named according to seq1 -- left side.
  Ograph s2rAs1(s2seq, len);     //s2 named according to seq1 -- right side.
  Ograph s2rAs2l(idstr, len);    //s2r named according to s2l.
  Ograph s2lAs2r(idstr, len);    //s2l named according to s2r.
  Ograph s2l = s2, s2r = s2;

  free(s1seq);
  free(s2seq);

          //Rotate the s2s around s1:
          //(naming as though s2 is at the top and s1 at the bottom)
  Pair inv;
  for(int i=0; i<dist/2; i++)
    {                              //Move left side:
    inv = getRotatingInv(s2lAs2, s2lAs2r, s2lAs1); //Get an inversion
    cerr << "(" << s2rAs1.getDistNH() << "," << s2lAs1.getDistNH() 
         << "," << s2rAs2l.getDistNH() << ")";
    if(inv != Pair(0,0))                           //and do it:
      {
      s2lAs1.revAll(inv);
      s2lAs2.revAll(inv);
      s2lAs2r.revAll(inv);
      s2rAs2l.revAllInv(inv);
      s2l.revAll(inv);
      }
    else
      cerr << "NULL:" << i <<  endl;

    cerr << "(" << s2rAs1.getDistNH() << "," << s2lAs1.getDistNH() 
         << "," << s2rAs2l.getDistNH() << ")";
                                   //Move right side:
    inv = getRotatingInv(s2rAs2, s2rAs2l, s2rAs1);
    if(inv != Pair(0,0))
      {
      s2rAs1.revAll(inv);
      s2rAs2.revAll(inv);
      s2rAs2l.revAll(inv);
      s2lAs2r.revAllInv(inv);
      s2r.revAll(inv);
      }
    else
      cerr << "NULL:" << i <<  endl;

    cerr << "(" << s2rAs1.getDistNH() << "," << s2lAs1.getDistNH() 
         << "," << s2rAs2l.getDistNH() << ")" << ":"
         << s2r.getDistNH() << "-" << s2l.getDistNH() << endl;
    }

  }


void method1(int len, int dist)
  {
  int* idstr = makeID(len);      //Build the identity sequence.
  Ograph s1(idstr, len);         //Build the identity for seq1.
  Ograph s2(idstr, len);         //Build the identity for seq2.

          //Make s1 be dist/2 from the ID:
  for(int i=0; i < dist/2;i++)
    {
    s1.revAll(1,1+i);
    s2.revAll(len-i,len);
    }

          //Set up the graphs for s2 and s1 renamed according to each:
  int* s1seq = s1.getSeq();
  int* s2seq = s2.getSeq();

  Ograph s1As1(idstr, len);      //s1 named according to seq1.

  Ograph s2lAs2(idstr, len);     //s2 named according to seq2 -- left side.
  Ograph s2rAs2(idstr, len);     //s2 named according to seq2 -- right side.
  seqACCseq(s1seq, s2seq, len);  //Get s2 named as though s1 were the ID.
  Ograph s2lAs1(s2seq, len);     //s2 named according to seq1 -- left side.
  Ograph s2rAs1(s2seq, len);     //s2 named according to seq1 -- right side.
  Ograph s2rAs2l(idstr, len);    //s2r named according to s2l.
  Ograph s2lAs2r(idstr, len);    //s2l named according to s2r.
  Ograph s2l = s2, s2r = s2;

  free(s1seq);
  free(s2seq);

          //Rotate the s2s around s1:
          //(naming as though s2 is at the top and s1 at the bottom)
  Pair inv;
  for(int i=0; i<dist/2; i++)
    {                              //Move left side:
    inv = getRotatingInv(s2lAs2, s2lAs2r, s2lAs1); //Get an inversion
    cerr << "(" << s2rAs1.getDistNH() << "," << s2lAs1.getDistNH() 
         << "," << s2rAs2l.getDistNH() << ")";
    if(inv != Pair(0,0))                           //and do it:
      {
      s2lAs1.revAll(inv);
      s2lAs2.revAll(inv);
      s2lAs2r.revAll(inv);
      s2rAs2l.revAllInv(inv);
      s2l.revAll(inv);
      }
    else
      cerr << "NULL:" << i <<  endl;

    cerr << "(" << s2rAs1.getDistNH() << "," << s2lAs1.getDistNH() 
         << "," << s2rAs2l.getDistNH() << ")";
                                   //Move right side:
    inv = getRotatingInv(s2rAs2, s2rAs2l, s2rAs1);
    if(inv != Pair(0,0))
      {
      s2rAs1.revAll(inv);
      s2rAs2.revAll(inv);
      s2rAs2l.revAll(inv);
      s2lAs2r.revAllInv(inv);
      s2r.revAll(inv);
      }
    else
      cerr << "NULL:" << i <<  endl;

    cerr << "(" << s2rAs1.getDistNH() << "," << s2lAs1.getDistNH() 
         << "," << s2rAs2l.getDistNH() << ")" << ":"
         << s2r.getDistNH() << "-" << s2l.getDistNH() << endl;
    }

  Ograph s1lAs1(idstr, len);
  Ograph s1rAs1(idstr, len);
  Ograph s1lAs2l = s2lAs1.getInvGraph();
  Ograph s1rAs2r = s2rAs1.getInvGraph();
  Ograph s1lAs1r(idstr, len), s1rAs1l(idstr, len);
  Ograph s1l = s1, s1r = s1;
  for(int i=0; i<dist/2; i++)
    {
    inv = getRotatingInv(s1lAs1, s1lAs1r, s1lAs2l);
    if(inv != Pair(0,0))
      {
      s1lAs1.revAll(inv);
      s1lAs1r.revAll(inv);
      s1rAs1l.revAllInv(inv);
      s1lAs2l.revAll(inv);
      s1l.revAll(inv);
      }

    inv = getRotatingInv(s1rAs1, s1rAs1l, s1rAs2r);
    if(inv != Pair(0,0))
      {
      s1rAs1.revAll(inv);
      s1rAs1l.revAll(inv);
      s1lAs1r.revAllInv(inv);
      s1rAs2r.revAll(inv);
      s1r.revAll(inv);
      }

    cerr << "(" << s1lAs1.getDistNH() << "," << s1lAs1r.getDistNH() 
         << "," << s1lAs2l.getDistNH() << ")" << ":"
         << s1l.getDistNH() << "-" << s1r.getDistNH() << endl;
    }

  int *s2lseq = s2l.getSeq();
  int *s2rseq = s2r.getSeq();
  int *s1lseq = s1l.getSeq();
  int *s1rseq = s1r.getSeq();
  int s2lvs2r = multiMED(s2lseq, len, s2rseq, len); //Top
  int s2lvs1l = multiMED(s2lseq, len, s1lseq, len); //Left
  int s1rvs1l = multiMED(s1rseq, len, s1lseq, len); //Bottom
  int s1rvs2r = multiMED(s1rseq, len, s2rseq, len); //Right
  int s1rvs2l = multiMED(s1rseq, len, s2lseq, len); //Diagonal
  int s1lvs2r = multiMED(s1lseq, len, s2rseq, len); //Diagonal
  cerr << endl << s2lvs2r << "(top) =" << s2lvs1l << "(left)" << endl;
  cerr << s1rvs1l << "(bottom) =" << s1rvs2r << "(right)" << endl;
  cerr << "diagonals:" << s1rvs2l << "," << s1lvs2r << endl;
  cerr << "s2l:"; s2l.printSeqStr(); cerr << endl;
  cerr << "s2r:"; s2r.printSeqStr(); cerr << endl;
  cerr << "s1l:"; s1l.printSeqStr(); cerr << endl;
  cerr << "s1r:"; s1r.printSeqStr(); cerr << endl;
  assert(s2lvs2r == s2lvs1l);
  assert(s1rvs1l == s1rvs2r);

  Ograph tid(idstr, len);                          //Build the identity.
  Ograph lid=tid,bid=tid,rid=tid;
  Ograph dlr=tid,drl=tid;
  cerr << "getting top sig..." << endl;
  int topsigsize = getSigSize(s2l, s2r, tid);
  cerr << "\t" << topsigsize << endl;
  cerr << "getting left sig..." << endl;
  int leftsigsize = getSigSize(s2l, s1l, lid);
  cerr << "\t" << leftsigsize << endl;
  cerr << "getting bottom sig..." << endl;
  int bottomsigsize = getSigSize(s1r, s1l, bid);
  cerr << "\t" << bottomsigsize << endl;
  cerr << "getting right sig..." << endl;
  int rightsigsize = getSigSize(s1r, s2r, rid);
  cerr << "\t" << rightsigsize << endl;

  cerr << "getting diagonal(l->r) sig..." << endl;
  int dlrsigsize = getSigSize(s2l, s1r, tid);
  cerr << "\t" << dlrsigsize << endl;
  cerr << "getting diagonal(r->l) sig..." << endl;
  int drlsigsize = getSigSize(s2r, s1l, tid);
  cerr << "\t" << drlsigsize << endl;


  free(idstr);
  }


/* Return a Pair that represents an inversion that will rotate around the
implicit identity given by -around- and moves away from -from-.
*/
Pair getRotatingInv(const Ograph &from, const Ograph &around)
  {
  vector<Pair> un, ben;

  from.getUnsortingInvs(un);
  around.getBenignInvs(ben);

  for(int i=0; size_t(i)<un.size(); i++)
    for(int j=0; size_t(j)<ben.size(); j++)
      if(un[i] == ben[j])
         return un[i];

  return Pair(0,0);
  }

/* Return a Pair that represents an inversion that will rotate around the
implicit identity given by -around- and moves away from -from1- and -from2-.
*/
Pair getRotatingInv(const Ograph &from1, const Ograph &from2,
                    const Ograph &around)
  {
  vector<Pair> un1, un2, ben;

  from1.getUnsortingInvs(un1);
  from2.getUnsortingInvs(un2);
  around.getBenignInvs(ben);

  for(int i=0; size_t(i)<un1.size(); i++)
    for(int k=0; size_t(k)<un2.size(); k++)
      for(int j=0; size_t(j)<ben.size(); j++)
        if((un1[i] == ben[j]) && (un2[k] == ben[j]))
          return un1[i];

  return Pair(0,0);
  }
