/* Krister Swenson                                                 Winter 2007-8

A vertex that knows it's neighbors and has a number and string label, and can
be marked as visited.
*/


#include "Vertex.h"

using namespace std;

Vertex::Vertex()
  { label=0; degree=0; themark=false; name=""; }
Vertex::Vertex(int l)
  {
  label=l;
  stringstream out;
  out << l;
  name = out.str();

  degree=0;
  themark=false;
  }
Vertex::Vertex(int l, string n)
  {
  label=l;
  name = n;
  degree=0;
  themark=false;
  }
Vertex::Vertex(const Vertex &v)
  { copy(v); }
Vertex& Vertex::operator=(const Vertex &v)
  { copy(v); return *this; }
Vertex::~Vertex()
  { }
bool Vertex::operator!=(const Vertex &v) const
  { return !(*this == v); }
bool Vertex::operator==(const Vertex &v) const
  {
  if(v.label != label)
    return false;
  for(vpmap::const_iterator i = v.edgemap.begin(); i != v.edgemap.end(); i++)
    if(edgemap.find(i->second->getLabel()) == edgemap.end())
      return false;  //Return false if the vertex is not a neighbor of this.
  for(vpmap::const_iterator i = edgemap.begin(); i != edgemap.end(); i++)
    if(v.edgemap.find(i->second->getLabel()) == v.edgemap.end())
      return false;  //Return false if the vertex is not a neighbor of this.
  return true;
  }
int Vertex::getLabel() const
  { return label; }
string Vertex::getName() const
  { return name; }
void Vertex::mark()
  { themark=true; }
void Vertex::unmark()
  { themark=false; }
bool Vertex::getMark() const
  { return themark; }

/*Return true if none of this vertex's neighbors are marked.
*/
bool Vertex::noMarkedNeighbor() const
  {
  for(vpmap::const_iterator i = edgemap.begin(); i != edgemap.end(); i++)
    if(i->second->getMark())
      return false;
  return true;
  }

/*Return the begin iterator to visit all neighbors with.
*/
vpmap::const_iterator Vertex::nBeg() const
  { return edgemap.begin(); }
/*Return the end iterator to visit all neighbors with.
*/
vpmap::const_iterator Vertex::nEnd() const
  { return edgemap.end(); }
/*Return true if the given vertex is a neighbor of this one.
*/
bool Vertex::hasNeighbor(const Vertex *v) const
  { return edgemap.find(v->getLabel()) != edgemap.end(); }
bool Vertex::hasNeighbor(int l) const
  { return edgemap.find(l) != edgemap.end(); }

/*Return true if a vertex from -l- is a neighbor of this one.
*/
bool Vertex::hasNeighbor(const list<int> &l) const
  {
  for(list<int>::const_iterator i = l.begin(); i != l.end(); i++)
    if(hasNeighbor(*i))
      return true;

  return false;
  }



int Vertex::getDegree() const
  { return degree; }

/*Make -v- a neighbor of this vertex if it already isn't.
*/
void Vertex::addNeighbor(Vertex *v)
  {
  if(edgemap.find(v->label) == edgemap.end())
    {
    edgemap[v->label] = v;
    degree++;
    }
  }

/*Remove -v- as a neighbor of this vertex if it already isn't.
*/
void Vertex::remNeighbor(Vertex *v)
  {
  vpmap::iterator i;
  if((i = edgemap.find(v->getLabel())) != edgemap.end())
    {
    degree--;
    edgemap.erase(i);
    }
  }

void Vertex::copy(const Vertex &v)
  {
  label=v.label; degree=v.degree; themark=v.themark;
  edgemap = v.edgemap;
  name = v.name;
  }

ostream& operator<<(ostream& output, const Vertex& v)
  {
  output << v.label << ":" << v.edgemap;
  return output;
  }


ostream& operator<<(ostream& output, const vpmap& x)
  {
  for(vpmap::const_iterator i = x.begin(); i != x.end(); i++)
    output << i->second->getLabel() << " ";
  return output;
  }
